@@ -3,9 +3,12 @@
 from utz import *
 
 from .cli import Arg, run_args, load_run_config
-from .config import clean_group, clean_mount, lists, Config, DEFAULT_IMAGE_REPO, DEFAULT_SRC_MOUNT_DIR, DEFAULT_RUN_NB, IMAGE_HOME, DEFAULT_IMAGE, DEFAULT_DIND_IMAGE
+from .config import clean_group, lists, version, Config, DEFAULT_IMAGE_REPO, DEFAULT_SRC_DIR_NAME, DEFAULT_SRC_MOUNT_DIR, DEFAULT_RUN_NB, IMAGE_HOME, DEFAULT_GROUP, DEFAULT_USER, DEFAULT_IMAGE, DEFAULT_DIND_IMAGE, GSMO_DIR, GSMO_DIR_NAME
+from .err import OK, RAISE, WARN
+from .mount import Mount, Mounts
 
-def main():
+def main(*args):
+    print(f'gsmo.main({args})')
     parser = ArgumentParser()
     parser.add_argument('input',nargs='?',help='Input directory containing run.ipynb (and optionally gsmo.yml, or other path specified by "-y"); defaults to current directory')
 
@@ -20,6 +23,7 @@ def main():
     docker_args = [
         Arg('-a','--apt',help='Comma-separated list of packages to apt-get install'),
         Arg('-b','--build-arg',action='append',help='Comma-separated list of packages to apt-get install'),
+        Arg('--dev',default=None,action='store_true',help="Run in dev mode: use a 'latest' Docker image tag (':latest' or ':dind') and mount this gsmo directory into the Docker image (as /gsmo)"),
         Arg('--dind',default=None,action='store_true',help="When set, mount /var/run/docker.sock in container (and default to a base image that contains docker installed)"),
         Arg('--dst',help='Path inside Docker container to mount current directory/repo to (default: /src)'),
         Arg('-e','--env','--image-env',action='append',help='Environment variables to set in Docker image (at build time)'),
@@ -27,20 +31,23 @@ def main():
         Arg('-E','--container-env',action='append',help='Environment variables to pass to Docker container (at run time)'),
         Arg('--container-env-file','--Ef',action='append',help='Files containing environment variables to pass to Docker container (at run time)'),
         Arg('-i','--image',help=f'Base docker image to build on (default: f{DEFAULT_IMAGE})'),
+        Arg('--id',help='Comma-delimited subset of {user,group,root,sudo} (or {u,g,r,s}); short-hand for --image-user, --image-group, --root, and --sudo, resp.'),
         Arg('-I','--no-interactive',default=None,action='store_true',help="Don't run interactively / allocate a TTY (i.e. skip `-it` flags to `docker run`)"),
-        Arg('-g','--image-group',default=None,action='store_true',help='Create current group inside Docker image (at build time)'),
-        Arg('-G','--group',action='append',help='Groups to add user to when running the Docker container'),
+        Arg('-g','--image-group',help='Create a group with this name inside the Docker image (with same GID as the current host-machine group; empty string implies using the current host-machine group name)'),
+        Arg('-G','--group',action='append',help="Additional groups to add docker image user to"),
+        Arg('-l','--label',action='append',help='Labels to apply to run container, in k=v format'),
+        Arg('-L','--label-file',help='File with labels to apply to run container, in k=v format'),
         Arg('-M','--missing-paths',default=0,action='count',help='Relax checking of paths (for propagating mounts and groups into Docker): 1x ⟹ warn, 2x ⟹ ignore'),
         Arg('-n','--dry-run',action='count',default=0,help="Prepare and print run cmd (including building Docker image), but don't execute it. If passed twice, stop before building Docker image"),
         Arg('--name',help='Container name (defaults to directory basename)'),
         Arg('-p','--pip',help='Comma-separated (or multi-arg) list of packages to pip install'),
-        Arg('--container-pip','--pie',action='append',help='When running the container, `pip install -e` a directory or directories (especially subdirectories of the project being run, which are mounted into the container and are not available for `pip install`ing at image-build time) before running the usual entrypoint script'),
+        Arg('--container-pip','--pie','--pip-e',action='append',help='When running the container, `pip install -e` a directory or directories (especially subdirectories of the project being run, which are mounted into the container and are not available for `pip install`ing at image-build time) before running the usual entrypoint script'),
         Arg('-P','--port',action='append',help='Ports (or ranges) to expose from the container (if Jupyter server is being run, the first port in the first provided range will be used); can be passed multiple times and/or as comma-delimited lists'),
         Arg('--rm','--remove-container',default=None,action='store_true',help="Remove Docker container after run (pass `--rm` to `docker run`)"),
         Arg('-R','--skip-requirements-txt',default=None,action='store_true',help="Skip {reading,`pip install`ing} any requirements.txt that is present"),
         Arg('--sudo',default=None,action='store_true',help="Ensure Docker image user has sudo privileges"),
         Arg('-t','--tag',help='Comma-separated (or multi-arg) list of tags to add to built docker image'),
-        Arg('-u','--image-user',default=None,action='store_true',help="Create current user inside Docker image (at build time)"),
+        Arg('-u','--image-user',help='Create a user with this name inside the Docker image (with same UID as the current host-machine user; empty string implies using the current host-machine user name)'),
         Arg('-U','--root','--no-user',default=None,action='store_true',help="Run docker as root (instead of as the current system user)"),
         Arg('-v','--mount',action='append',help='Paths to mount into Docker container; relative paths are accepted, and the destination can be omitted if it matches the src (relative to the current directory, e.g. "home" → "/home")'),
     ]
@@ -65,7 +72,11 @@ def main():
     for arg in run_args:
         run_parser.add_argument(*arg.args, **arg.kwargs)
 
-    args = parser.parse_args()
+    if args:
+        args = parser.parse_args(args)
+    else:
+        args = parser.parse_args()
+
     jupyter_mode = shell_mode = run_mode = False
     cmd = getattr(args, 'cmd', None)
     if cmd =='jupyter':
@@ -94,7 +105,37 @@ def main():
     config = Config(args)
     get = partial(Config.get, config)
 
-    dst = get('dst',DEFAULT_SRC_MOUNT_DIR)
+    container_pips = lists(get('container_pip')) + lists(get('pie'))
+    pips = get('pip', [])
+    if isinstance(pips, dict):
+        keys = pips.keys()
+        if 'container' in keys: container_pips += pips.pop('container')
+        if 'img' in keys:
+            assert 'image' not in keys
+            pips = pips.pop('img')
+        elif 'image' in keys:
+            pips = pips.pop('image')
+        if keys:
+            raise ValueError(f'Unexpected keys in `pip` config dict (expected: "container" || ("img" ^ "image")): {keys}')
+
+    gsmo_root = env.get('GSMO_ROOT')
+    dst = get('dst', env.get('GSMO_DST'))
+    gsmo_dir = env.get('GSMO_DIR')
+    if gsmo_root:
+        if not gsmo_dir:
+            gsmo_dir = join(gsmo_root, GSMO_DIR_NAME)
+        if dst:
+            if not abspath(dst):
+                dst = join(gsmo_root, dst)
+        else:
+            dst = join(gsmo_root, DEFAULT_SRC_DIR_NAME)
+    else:
+        gsmo_root = '/'
+        if not gsmo_dir:
+            gsmo_dir = GSMO_DIR
+        if not dst:
+            dst = DEFAULT_SRC_MOUNT_DIR
+    print(f'gsmo_dir: {gsmo_dir}, dst {dst} (root {gsmo_root})')
     src = cwd
 
     jupyter_dir = get('dir') or dst
@@ -137,6 +178,18 @@ def main():
 
     image_env_file = get('env_file')
 
+    # Load container labels
+    labels = get('label', [])
+    if isinstance(labels, (list, tuple)):
+        labels = dict([
+            env.split('=', 1)
+            for env in labels
+        ])
+    elif labels is not None and not isinstance(labels, dict):
+        raise ValueError(f'Unexpected env dict: {labels}')
+
+    labels_file = get('label_file')
+
     # Load env var configs
     container_envs = get('container_env', [])
     if isinstance(container_envs, (list, tuple)):
@@ -151,13 +204,15 @@ def main():
 
     commit = lists(get('commit'))
 
+    dev_mode = get('dev', env.get('GSMO_DEV_MODE', False))
+
     missing_paths = get('missing_paths')
     if missing_paths == 1:
-        from .err import WARN
         missing_paths = WARN
     elif missing_paths == 2:
-        from .err import OK
         missing_paths = OK
+    else:
+        missing_paths = RAISE
 
     groups = lists(get('group'))
     groups = [ g for group in groups if (g := clean_group(group, err=missing_paths)) ]
@@ -165,34 +220,100 @@ def main():
     out = get('out') or 'nbs'
 
     mounts = lists(get('mount', []))
-    mounts = [ m for mount in mounts if (m := clean_mount(mount, err=missing_paths)) ]
-    mounts += [ f'{src}:{dst}', ]
+    mounts = Mounts(mounts, err=missing_paths)
+    env_mnts = env.get('GSMO_MOUNTS')
+    if env_mnts:
+        env_mnts = Mounts(env_mnts, keep_missing=True)
+
+    def dind_mnt(src, dst):
+        mnt = Mount(src, dst, err=missing_paths)
+        print(f'inspecting mount {mnt} for re-mapping: {env_mnts}')
+        if env_mnts:
+            dst2src = env_mnts.dst2src
+            dir = src
+            relpath = None
+            while True:
+                if dir in dst2src:
+                    host_src = dst2src[dir]
+                    if relpath:
+                        host_src = join(host_src, relpath)
+                    host_mnt = Mount(host_src,dst,keep_missing=True)
+                    print(f'Re-mapping mount {mnt} to host src: {host_mnt}')
+                    return host_mnt
+                parent = dirname(dir)
+                if parent == dir:
+                    break
+                if not relpath:
+                    relpath = basename(dir)
+                else:
+                    relpath = join(basename(dir), relpath)
+                dir = parent
+        return mnt
+
+    mounts = Mounts([ dind_mnt(m.src, m.dst) for m in mounts.mounts ])
+    mounts += dind_mnt(src, dst)
 
     dind = get('dind')
     if dind:
         default_image = DEFAULT_DIND_IMAGE
-        mounts += [ '/var/run/docker.sock:/var/run/docker.sock' ]
+        mounts += Mount('/var/run/docker.sock', err=RAISE)
     else:
         default_image = DEFAULT_IMAGE
-    base_image = get('image', default_image)
+    base_image = get('image')
+    if base_image:
+        explict_base_img = True
+    else:
+        explict_base_img = False
+        base_image = default_image
     if base_image.startswith(':'):
-        # shorthand for just specifying a runsascoded/gsmo tag
-        base_image = f'{DEFAULT_IMAGE_REPO}{base_image}'
+        if base_image == ':' or base_image == ':dind':
+            if dind:
+                base_image = ':dind'
+            if base_image == ':':
+                base_image = DEFAULT_IMAGE_REPO
+            else:
+                base_image = f'{DEFAULT_IMAGE_REPO}{base_image}'
+            dev_mode = True
+        else:
+            # shorthand for just specifying a runsascoded/gsmo tag
+            base_image = f'{DEFAULT_IMAGE_REPO}{base_image}'
     image = base_image
 
+    if dev_mode:
+        gsmo_dev_dir = dirname(dirname(__file__))
+        gsmo_mount = dind_mnt(gsmo_dev_dir, gsmo_dir)
+        print(f'Adding gsmo mount and container editable-pip: {gsmo_mount}')
+        mounts += gsmo_mount
+        container_pips += [gsmo_dir]
+
     use_docker = get('docker', True)
     rm = get('remove_container')
 
     ports = lists(get('port'))
     apts = lists(get('apt'))
-    pips = lists(get('pip'))
+
     tags = lists(get('tag'))
-    name = get('name', default=basename(cwd))
+    name = get('name', default=basename(cwd)).lower()
     skip_requirements_txt = args.skip_requirements_txt
     root = get('root')
-    image_user = get('image_user')
-    image_group = get('image_group')
+
+    from .util.unix_id import UnixId
+    id = UnixId()
+
+    image_user = get('image_user', DEFAULT_USER)
+    if image_user == '': image_user = id.user
+    image_group = get('image_group', DEFAULT_GROUP)
+    if image_group == '': image_group = id.group
     sudo = get('sudo')
+    id_attrs = lists(get('id'))
+    if 'u' in id_attrs or 'user' in id_attrs: image_user = id.user
+    if 'g' in id_attrs or 'group' in id_attrs: image_group = id.group
+    if 'r' in id_attrs or 'root' in id_attrs: root = True
+    if 's' in id_attrs or 'sudo' in id_attrs: sudo = True
+
+    if container_pips:
+        sudo = True
+
     dry_run = get('dry_run')
 
     if jupyter_mode:
@@ -251,8 +372,16 @@ def main():
         else:
             ports = []
 
-    from .util.unix_id import UnixId
-    id = UnixId()
+    if dind:
+        [gid,grp] = line(
+            'docker','run',
+            '-v','/var/run/docker.sock:/var/run/docker.sock',
+            '--rm','--entrypoint','stat',
+            base_image,
+            '-c','%g %G','/var/run/docker.sock',
+        ).split(' ')
+        docker_sock = o(gid=gid,grp=grp)
+        print(f'Parsed /var/run/docker.sock group: {grp} ({gid})')
 
     from utz import docker
     from utz.use import use
@@ -261,9 +390,8 @@ def main():
     # based from it; otherwise, use an extant upstream image
     build_image = False
 
-    # docker_gid = None
     dockerfile = join(cwd, 'Dockerfile')
-    if exists(dockerfile):
+    if exists(dockerfile) and not explict_base_img:
         build_image = True
         extend = dockerfile
     else:
@@ -287,7 +415,11 @@ def main():
         reqs_txt = join(cwd, 'requirements.txt')
         if exists(reqs_txt) and not skip_requirements_txt:
             with open(reqs_txt, 'r') as f:
-                pips += [ line.rstrip('\n') for line in f.readlines() if line ]
+                pips += [
+                    f'"{dep}"'
+                    for line in f.readlines()
+                    if (dep := line.rstrip('\n'))
+                ]
 
         if pips:
             if use_docker:
@@ -298,33 +430,63 @@ def main():
                 print(f'pip install {" ".join(pips)}')
                 pip.main(['install'] + pips)
 
+        default_kvs = {
+            'cmd': cmd,
+            'dir': gsmo_dir,
+            'dev_mode': dev_mode,
+            'dst': dst,
+            'image': base_image,
+            'mounts': str(mounts),
+            'path': cwd,
+            'root': gsmo_root,
+            'version': version,
+        }
+
+        ENV({ f'GSMO_{k.upper()}':v for k,v in default_kvs.items()})
+
         if image_envs:
             build_image = True
-            ENV(**image_envs)
+            ENV(image_envs)
 
         if image_env_file:
             build_image = True
             with open(image_env_file,'r') as f:
                 ENV(*[ l.strip() for l in f.readlines() ])
 
+        LABEL({ f'gsmo.{k}':v for k,v in default_kvs.items()})
+
+        if labels:
+            build_image = True
+            LABEL(**labels)
+
+        if labels_file:
+            build_image = True
+            with open(labels_file,'r') as f:
+                LABEL(*[ l.strip() for l in f.readlines() ])
+
         if use_docker:
-            if image_user or image_group or sudo:
+            if image_user or image_group or sudo or dind:
                 cmds = []
-                if image_group: cmds += [f'groupadd -f -o -g {id.gid} {id.group}']
-                if image_user:
+
+                if image_group or dind:
+                    assert image_group
+                    cmds += [f'groupadd -f -o -g {id.gid} {image_group}']
+
+                if image_user or dind:
+                    assert image_user
                     if dind:
-                        useradd = f'useradd -u {id.uid} -g {id.gid} -G docker -s /bin/bash -m -d {IMAGE_HOME} {id.user}'
+                        useradd = f'useradd -u {id.uid} -g {id.gid} -G {docker_sock.gid} -s /bin/bash -m -d {IMAGE_HOME} {image_user}'
                     else:
-                        useradd = f'useradd -u {id.uid} -g {id.gid} -s /bin/bash -m -d {IMAGE_HOME} {id.user}'
+                        useradd = f'useradd -u {id.uid} -g {id.gid} -s /bin/bash -m -d {IMAGE_HOME} {image_user}'
                     cmds += [
                         useradd,
                         f'chown -R {id.uid}:{id.gid} {IMAGE_HOME}',
                     ]
-                if sudo: cmds += [
-                    'apt-get update',
-                    'apt-get install -y sudo',
-                    'perl -pi -e "s/^%%sudo(.*ALL=).*/%s\\1(ALL) NOPASSWD: ALL/" /etc/sudoers' % id.user,
-                ]
+
+                if sudo or dind:
+                    # user isn't known at build-time though, so pswd-less sudo is patched in here
+                    cmds += [ 'perl -pi -e "s/^%%sudo(.*ALL=).*/%s\\1(ALL) NOPASSWD: ALL/" /etc/sudoers' % image_user, ]
+
                 build_image = True
                 RUN(*cmds)
                 if image_user:
@@ -359,9 +521,16 @@ def main():
             user_args = [ '-u', f'{uid}:{gid}' ]
 
     # Remove any existing container
+    run_in_existing_container = False
     if use_docker:
-        if check('docker','container','inspect',name):
-            run('docker','container','rm',name)
+        container = process.json('docker','container','inspect',name, err_ok=True)
+        if container:
+            container = singleton(container, dedupe=False)
+            if container.get('State',{}).get('Running',False):
+                print(f'Will execute in existing container {name}')
+                run_in_existing_container = True
+            else:
+                run('docker','container','rm',name)
 
     interactive = not args.no_interactive
     if interactive:
@@ -370,7 +539,8 @@ def main():
         flags = []
     if rm:
         assert use_docker
-        flags += ['--rm']
+        if not run_in_existing_container:
+            flags += ['--rm']
 
     run_nb = get('run', DEFAULT_RUN_NB)
 
@@ -399,19 +569,29 @@ def main():
         if commit:
             args += [ ['--commit',path] for path in commit]
 
-    container_pips = lists(get('container_pip'))
     if container_pips:
         args = [ len(container_pips) ] + container_pips + [ entrypoint ] + args
-        entrypoint = '/gsmo/pip_entrypoint.sh'
+        entrypoint = join(gsmo_dir,'pip_entrypoint.sh')
+
+    if dind:
+        args = [entrypoint] + args
+        entrypoint = join(gsmo_dir,'dind_entrypoint.sh')
+        groups.append(docker_sock.gid)
 
     if run_mode:
         RUN_CONFIG_YML_PATH = '/run_config.yml'
         if run_config:
-            run_config_path = NamedTemporaryFile(delete=False)
-            with open(run_config_path.name,'w') as f:
-                yaml.safe_dump(dict(run_config), f)
-            mounts += [ f'{run_config_path.name}:{RUN_CONFIG_YML_PATH}' ]
+            run_config_file = NamedTemporaryFile(dir=env.get('GSMO_DIR'), suffix='.yml', delete=False)
+            run_config_path = run_config_file.name
+            print(f'Writing run config to {run_config_path} (gsmo dir: {gsmo_dir})')
+            with open(run_config_path,'w') as f:
+                yaml.safe_dump(dict(run_config), f, sort_keys=False)
+            mounts += dind_mnt(run_config_path, RUN_CONFIG_YML_PATH)
             args += [ '-Y',RUN_CONFIG_YML_PATH ]
+        else:
+            print('no run config')
+    else:
+        print('no run mode')
 
     def get_git_id(k, fmt):
         try:
@@ -422,7 +602,7 @@ def main():
         return v
 
     # Get Git user name/email for propagating into image
-    id = o(
+    git_id = o(
         name  = get_git_id( 'name', '%an'),
         email = get_git_id('email', '%ae'),
     )
@@ -430,56 +610,80 @@ def main():
     # Set up author info for git committing
     container_envs = {
         **container_envs,
-       'GIT_AUTHOR_NAME'    : id.name,
-       'GIT_AUTHOR_EMAIL'   : id.email,
-       'GIT_COMMITTER_NAME' : id.name,
-       'GIT_COMMITTER_EMAIL': id.email,
+       'GIT_AUTHOR_NAME'    : git_id.name,
+       'GIT_AUTHOR_EMAIL'   : git_id.email,
+       'GIT_COMMITTER_NAME' : git_id.name,
+       'GIT_COMMITTER_EMAIL': git_id.email,
     }
 
     # Build Docker CLI args
     env_args = [ [ '-e', f'{k}={v}' ] for k, v in container_envs.items() ]
     if container_env_file: env_args += [ '--env-file', container_env_file ]
-    mount_args = [ [ '-v', mount ] for mount in mounts ]
     port_args = [ [ '-p', port ] for port in ports ]
     group_args = [ [ '--group-add', group ] for group in groups ]
     entrypoint_args = [ '--entrypoint', entrypoint ]
     workdir_args = [ '--workdir', workdir ]
+    name_args = [ '--name', name ]
 
-    all_args = \
+    exec_flags = \
         flags + \
-        entrypoint_args + \
         env_args + \
-        mount_args + \
+        workdir_args
+
+    print(f'mounts: {mounts}')
+    all_flags = \
+        exec_flags + \
+        mounts.args() + \
         port_args + \
         user_args + \
-        workdir_args + \
-        group_args + \
-        [image] + \
-        args
+        group_args
+
+    if run_in_existing_container:
+        all_args = \
+            exec_flags + \
+            [name] + \
+            [entrypoint] + \
+            args
+    else:
+        all_args = \
+            all_flags + \
+            entrypoint_args + \
+            name_args + \
+            [image] + \
+            args
 
     if use_docker:
         if jupyter_mode and check('which', 'open'):
             # 1. run docker container in detached mode
             # 2. parse+open jupyter token URL in browser (try every 1s)
             # 3. re-attach container
-            cmd = [
-                'docker','run',
-                '-w',dst,
-                '-d',
-                '--name',name,
-                all_args,
-            ]
+            if run_in_existing_container:
+                cmd = [
+                    'docker','exec',
+                    '-d',
+                    all_args,
+                ]
+            else:
+                cmd = [
+                    'docker','run',
+                    '-d',
+                    all_args,
+                ]
             if dry_run:
                 run(*cmd, dry_run=True)
             else:
                 run(*cmd)
+                sleep_interval = 0.5
+                backoff_idx = 0
+                backoff_cutoff = 5
                 while True:
+                    sleep(sleep_interval)
                     lns = lines('docker','exec',name,'jupyter','notebook','list')
                     if lns[0] != 'Currently running servers:':
                         raise Exception('Unexpected `jupyter notebook list` output:\n\t%s' % "\n\t".join(lns))
                     if len(lns) == 2:
                         ln = lns[1]
-                        rgx = f'(?P<url>http://0\.0\.0\.0:(?P<port>\d+)/\?token=(?P<token>[0-9a-f]+)) :: {jupyter_dir}'
+                        rgx = f'(?P<url>http://0\\.0\\.0\\.0:(?P<port>\\d+)/\\?token=(?P<token>[0-9a-f]+)) :: {jupyter_dir}'
                         if not (m := match(rgx, ln)):
                             raise RuntimeError(f'Unrecognized notebook server line: {ln}')
                         if m['port'] != str(jupyter_dst_port):
@@ -498,17 +702,25 @@ def main():
                                 run('docker','attach',name)
                         break
                     else:
-                        SLEEP_INTERVAL = 1
-                        print(f'No Jupyter server found in container {name}; sleep {SLEEP_INTERVAL}s…')
-                        sleep(SLEEP_INTERVAL)
+                        print(f'No Jupyter server found in container {name}; sleep {sleep_interval}s…')
+                        backoff_idx += 1
+                        if backoff_idx == backoff_cutoff:
+                            backoff_idx = 0
+                            sleep_interval *= 1.6
         else:
-            run(
-                'docker','run',
-                '-w',dst,
-                '--name',name,
-                all_args,
-                dry_run=dry_run,
-            )
+            print(f'running from {cwd}')
+            if run_in_existing_container:
+                run(
+                    'docker','exec',
+                    all_args,
+                    dry_run=dry_run,
+                )
+            else:
+                run(
+                    'docker','run',
+                    all_args,
+                    dry_run=dry_run,
+                )
     else:
         if jupyter_src_port != jupyter_dst_port:
             raise ValueError(f'Mismatching jupyter ports in non-docker mode: {jupyter_src_port} != {jupyter_dst_port}')
